//                              -*- Mode: Verilog -*-
// Filename        : rgb2yuv.v
// Description     : RGB to YUV convertion

module rgb2yuv (
                // Inputs
                HCLK,
                HRESETn,
                HSELMYIP,
                HREADYIn,
                HTRANS,
                HSIZE,
                HWRITE,
                HWDATA,
                HADDR,
                // Outputs
                HREADYOut,
                HRESP,
                SDATAEN,
                SnWBYTE,
                SnOE,
                SnCE,
                HRDATA,
                nRGB2YUV_INT
                );
   
   // Inputs
   input         HCLK;      // system bus clock
   input         HRESETn;   // reset input (active low)
   input         HSELMYIP;  // AHB peripheral select
   input         HREADYIn;  // AHB ready input
   input [1:0]   HTRANS;    // AHB transfer type
   input [1:0]   HSIZE;     // AHB hsize
   input         HWRITE;    // AHB hwrite
   input [31:0]  HWDATA;    // AHB write data bus
   input [31:0]  HADDR;     // AHB address bus
   
   // Outputs
   output        HREADYOut; // AHB ready output to S->M mux
   output [1:0]  HRESP;     // AHB response
   output        SDATAEN;   // MYIP tristate enable
   output [3:0]  SnWBYTE;   // MYIP byte lane writes
   output        SnOE;      // MYIP output enable
   output        SnCE;      // MYIP chip enable
   output [31:0] HRDATA;    // AHB read data bus
   output        nRGB2YUV_INT;   // Interrupt lline
   
   // Inputs
   wire          HCLK;      // system bus clock
   wire          HRESETn;   // reset input (active low)
   wire          HSELMYIP;  // AHB peripheral select
   wire          HREADYIn;  // AHB ready input
   wire [1:0]    HTRANS;    // AHB transfer type
   wire [1:0]    HSIZE;     // AHB hsize
   wire          HWRITE;    // AHB hwrite
   wire [31:0]   HWDATA;    // AHB write data bus
   wire [31:0]   HADDR;     // AHB address bus
   
   // Outputs
   wire          HREADYOut; // AHB ready output to S->M mux
   wire [1:0]    HRESP;     // AHB response
   wire          SDATAEN;   // MYIP tristate enable
   wire [3:0]    SnWBYTE;   // MYIP byte lane writes
   wire          SnOE;      // MYIP output enable
   wire          SnCE;      // MYIP chip enable
   wire [31:0]   HRDATA;    // AHB read data bus
   

`define ST_IDLE          2'b00
`define ST_READ          2'b01
`define ST_WRITE         2'b10

// HTRANS transfer type signal encoding
`define TRN_IDLE         2'b00
`define TRN_BUSY         2'b01
`define TRN_NONSEQ       2'b10
`define TRN_SEQ          2'b11

// HRESP transfer response signal encoding
`define RSP_OKAY         2'b00
`define RSP_ERROR        2'b01
`define RSP_RETRY        2'b10
`define RSP_SPLIT        2'b11

   
   /*
    * variables
    */
   wire          Valid;
   wire [31:0]   in0;
   wire [31:0]   in1; 
   wire [31:0]   in2;
   wire [31:0]   in3; 
   wire [31:0]   in4;
   wire [31:0]   in5; 
   wire [31:0]   in6;
   wire [31:0]   in7; 
   wire [23:0]   yuvout;
   
   reg  [3:0]    k;
   wire [3:0]    next_k;
   
   reg [1:0]     NextState;
   reg [1:0]     CurrentState;

   reg [23:0]    inreg0;
   reg [23:0]    inreg1;
   reg [23:0]    inreg2;
   reg [23:0]    inreg3;
   reg [23:0]    inreg4;
   reg [23:0]    inreg5;
   reg [23:0]    inreg6;
   reg [23:0]    inreg7;
   
   reg [23:0]    resultreg0;
   reg [23:0]    resultreg1;
   reg [23:0]    resultreg2;
   reg [23:0]    resultreg3;
   reg [23:0]    resultreg4;
   reg [23:0]    resultreg5;
   reg [23:0]    resultreg6;
   reg [23:0]    resultreg7;

   reg [23:0]    outreg;   
   reg [23:0]    next_outreg;
   
   reg [23:0]    rgbin; 
   reg           rgb_enable;
   wire          next_rgb_enable;
   
   wire          update_rgb_enable;
   wire          update_rgb_enable_int;
   wire          update_rgb_enable_ext;
   
   wire          clear_int;   
   wire          update_int;
   wire          next_nRGB2YUV_INT;
   wire          done;
   
   reg           nRGB2YUV_INT;  // the source of MYIP_INT interrupt signal

   reg [7:0]     write_reg;
   reg [7:0]     write_result;
   
   // -----------------------------------------------------------------------------
   // Next state logic for APB state machine
   // Generates next state from CurrentState and AHB inputs.
   // -----------------------------------------------------------------------------
   assign Valid = ((HSELMYIP == 1'b1) &
                   (HREADYIn == 1'b1) &
                   ((HTRANS == `TRN_NONSEQ) | (HTRANS == `TRN_SEQ))) ? 1'b1 : 1'b0;
   
   always @ (CurrentState or Valid or HWRITE)
     begin: p_NextStateComb
        case (CurrentState)
          // Idle state
          `ST_IDLE: begin
             if (Valid == 1'b1) begin
                if (HWRITE == 1'b1) begin
                   NextState <= `ST_WRITE;
                end
                else begin
                   NextState <= `ST_READ;
                end
             end
             else begin
                NextState <= `ST_IDLE;
             end
          end
          // second read cycle
          `ST_READ: begin
             if (Valid == 1'b1)begin
                if (HWRITE == 1'b1) begin
                   NextState <= `ST_WRITE;
                end 
                else begin
                   NextState <= `ST_READ;
                end
             end        
             else begin
                NextState <= `ST_IDLE;
             end
          end
          // second write cycle
          `ST_WRITE: begin
             if (Valid == 1'b1)begin
                if (HWRITE == 1'b1) begin
                   NextState <= `ST_WRITE;
                end   
                else begin
                   NextState <= `ST_READ;
                end
             end           
             else begin
                NextState <= `ST_IDLE;
             end
          end
          // Return to idle on FSM error
          default: begin
             NextState <= `ST_IDLE;
          end
        endcase
     end
   
   always @(negedge HRESETn or posedge HCLK)   
     begin : p_CurrentStSeq
        if (HRESETn == 1'b0) begin
           CurrentState <= `ST_IDLE;
        end
        else begin
           CurrentState <= NextState;
        end
     end
  
   /*
    * ZBT SSRAM output drivers
    */
   assign SnWBYTE = 4'b1111;
   assign SnCE    = 1'b1;
   assign SnOE    = 1'b1;
   assign SDATAEN = 1'b0;
   
   // -----------------------------------------------------------------------------
   // AHB output drivers
   // -----------------------------------------------------------------------------   
   assign HRESP = `RSP_OKAY;
   assign HREADYOut = 1'b1;
   assign HRDATA = {8'b0, outreg};
        
   /*
    * convertion core module
    */
   rgb2yuv_core u_rgb2yuv_core_1(HCLK, HRESETn, rgb_enable, rgbin, yuvout);

   /*
    * rgb2yuv decoder
    */
   always @(HWRITE or HADDR) begin
      write_reg = 8'b00000000;      
      
      if(HWRITE == 1'b1) begin
         case(HADDR[7:0])       
           8'h00: write_reg = 8'b00000001; 
           8'h04: write_reg = 8'b00000010;
           8'h08: write_reg = 8'b00000100;
           8'h0c: write_reg = 8'b00001000;
           8'h10: write_reg = 8'b00010000;
           8'h14: write_reg = 8'b00100000;
           8'h18: write_reg = 8'b01000000;
           8'h1c: write_reg = 8'b10000000;           
         endcase
      end
   end
   
   /*
    * rgb2yuv input registers
    */
   always @(posedge HCLK or negedge HRESETn) begin
      if(HRESETn == 1'b0) begin
         inreg0 <= 24'b0;       
      end
      else if(write_reg[0]) begin
         inreg0 <= HWDATA[23:0];        
      end
   end

   always @(posedge HCLK or negedge HRESETn) begin
      if(HRESETn == 1'b0) begin
         inreg1 <= 24'b0;       
      end
      else if(write_reg[1]) begin
         inreg1 <= HWDATA[23:0];        
      end
   end

   always @(posedge HCLK or negedge HRESETn) begin
      if(HRESETn == 1'b0) begin
         inreg2 <= 24'b0;       
      end
      else if(write_reg[2]) begin
         inreg2 <= HWDATA[23:0];        
      end
   end

   always @(posedge HCLK or negedge HRESETn) begin
      if(HRESETn == 1'b0) begin
         inreg3 <= 24'b0;       
      end
      else if(write_reg[3]) begin
         inreg3 <= HWDATA[23:0];        
      end
   end

   always @(posedge HCLK or negedge HRESETn) begin
      if(HRESETn == 1'b0) begin
         inreg4 <= 24'b0;       
      end
      else if(write_reg[4]) begin
         inreg4 <= HWDATA[23:0];        
      end
   end

   always @(posedge HCLK or negedge HRESETn) begin
      if(HRESETn == 1'b0) begin
         inreg5 <= 24'b0;       
      end
      else if(write_reg[5]) begin
         inreg5 <= HWDATA[23:0];        
      end
   end

   always @(posedge HCLK or negedge HRESETn) begin
      if(HRESETn == 1'b0) begin
         inreg6 <= 24'b0;   
      end
      else if(write_reg[6]) begin
         inreg6 <= HWDATA[23:0];    
      end
   end

   always @(posedge HCLK or negedge HRESETn) begin
      if(HRESETn == 1'b0) begin
         inreg7 <= 24'b0;   
      end
      else if(write_reg[7]) begin
         inreg7 <= HWDATA[23:0];    
      end
   end   

   /*
    * rgb2yuv result registers
    */
   always @(posedge HCLK or negedge HRESETn) begin
      if(HRESETn == 1'b0) begin
         resultreg0 <= 24'b0;   
      end
      else if(write_result[0]) begin
         resultreg0 <= yuvout;  
      end
   end

   always @(posedge HCLK or negedge HRESETn) begin
      if(HRESETn == 1'b0) begin
         resultreg1 <= 24'b0;   
      end
      else if(write_result[1]) begin
         resultreg1 <= yuvout;  
      end
   end

   always @(posedge HCLK or negedge HRESETn) begin
      if(HRESETn == 1'b0) begin
         resultreg2 <= 24'b0;   
      end
      else if(write_result[2]) begin
         resultreg2 <= yuvout;  
      end
   end

   always @(posedge HCLK or negedge HRESETn) begin
      if(HRESETn == 1'b0) begin
         resultreg3 <= 24'b0;   
      end
      else if(write_result[3]) begin
         resultreg3 <= yuvout;  
      end
   end

   always @(posedge HCLK or negedge HRESETn) begin
      if(HRESETn == 1'b0) begin
         resultreg4 <= 24'b0;   
      end
      else if(write_result[4]) begin
         resultreg4 <= yuvout;  
      end
   end

   always @(posedge HCLK or negedge HRESETn) begin
      if(HRESETn == 1'b0) begin
         resultreg5 <= 24'b0;   
      end
      else if(write_result[5]) begin
         resultreg5 <= yuvout;  
      end
   end

   always @(posedge HCLK or negedge HRESETn) begin
      if(HRESETn == 1'b0) begin
         resultreg6 <= 24'b0;   
      end
      else if(write_result[6]) begin
         resultreg6 <= yuvout;  
      end
   end

   always @(posedge HCLK or negedge HRESETn) begin
      if(HRESETn == 1'b0) begin
         resultreg7 <= 24'b0;   
      end
      else if(write_result[7]) begin
         resultreg7 <= yuvout;  
      end
   end   
   
   /*
    * rgb2yuv output mux
    */
   always @(HADDR or
            resultreg0 or resultreg1 or resultreg2 or resultreg3 or
            resultreg4 or resultreg5 or resultreg6 or resultreg7) begin
      case(HADDR[7:0])
        8'h20:   next_outreg = resultreg0;
        8'h24:   next_outreg = resultreg1;
        8'h28:   next_outreg = resultreg2;
        8'h2c:   next_outreg = resultreg3;
        8'h30:   next_outreg = resultreg4;
        8'h34:   next_outreg = resultreg5;
        8'h38:   next_outreg = resultreg6;
        8'h3c:   next_outreg = resultreg7;
        default: next_outreg = 24'b0;
      endcase
   end
   
   always @ (posedge HCLK or negedge HRESETn) begin
      if(HRESETn == 1'b0) begin
         outreg <= 24'b0;
      end
      else begin
         outreg <= next_outreg;
      end
   end
   
   /*
    * rgb2yuv control
    */
   assign next_rgb_enable = update_rgb_enable_ext ? HWDATA[0] : 1'b0;   
   assign update_rgb_enable_ext = (HWRITE == 1'b1) && (HADDR[7:0] == 8'h40);
   assign update_rgb_enable_int = !nRGB2YUV_INT;
   assign update_rgb_enable = update_rgb_enable_ext || update_rgb_enable_int;
   
   always @ (posedge HCLK or negedge HRESETn) begin
      if(HRESETn == 1'b0) begin
         rgb_enable <= 1'b0;
      end
      else if(update_rgb_enable == 1'b1) begin
         rgb_enable <= next_rgb_enable;
      end
   end

   assign clear_int = (HWRITE == 1'b1) && (HADDR[7:0] == 8'h44);
   assign update_int = done || clear_int;   
   assign next_nRGB2YUV_INT = clear_int ? 1'b1 : 1'b0;
   
   always @ (posedge HCLK or negedge HRESETn) begin
      if(HRESETn == 1'b0) begin
         nRGB2YUV_INT <= 1'b1;   
      end
      else if(update_int) begin
         nRGB2YUV_INT <= next_nRGB2YUV_INT;      
      end
   end
   

   /*
    * rgb2yuv control fsm
    */
   assign done = (k == 4'd12);
   
   assign next_k = rgb_enable ? (k + 1'b1) : 4'd0;
   
   always @ (posedge HCLK or negedge HRESETn) begin
      if(HRESETn == 1'b0) begin
         k <= 4'd0;      
      end
      else begin
         k <= next_k;    
      end      
   end

   always @ (k or
             inreg0 or inreg1 or inreg2 or inreg3 or
             inreg4 or inreg5 or inreg6 or inreg7) begin
      case(k)
        4'd1:    rgbin = inreg0;
        4'd2:    rgbin = inreg1;
        4'd3:    rgbin = inreg2;
        4'd4:    rgbin = inreg3;
        4'd5:    rgbin = inreg4;
        4'd6:    rgbin = inreg5;
        4'd7:    rgbin = inreg6;
        4'd8:    rgbin = inreg7;
        default: rgbin = 24'b0; 
      endcase      
   end
   
   always @ (k) begin
      case(k)
        4'd4:    write_result = 8'b00000001;
        4'd5:    write_result = 8'b00000010;
        4'd6:    write_result = 8'b00000100;
        4'd7:    write_result = 8'b00001000;
        4'd8:    write_result = 8'b00010000;
        4'd9:    write_result = 8'b00100000;
        4'd10:   write_result = 8'b01000000; 
        4'd11:   write_result = 8'b10000000;
        default: write_result = 8'b00000000; 
      endcase      
   end

endmodule
