//                              -*- Mode: Verilog -*-
// Filename        : rgb2yuv_core.v
// Description     : core function of RGB to YUV convertion

module rgb2yuv_core(
                    //inputs
                    HCLK,
                    HRESETn,                
                    enable,
                    rgb,
                    //output
                    yuv);
   /*
    * inputs
    */
   input         HCLK;
   input         HRESETn;                       
   input         enable;
   input [23:0]  rgb;

   /*
    * output
    */
   output [23:0] yuv;

   /*
    * variables
    */
   reg  [23:0]   yuv;
   wire [23:0]   yuv_out;
   
   wire [18:0]   y_r;
   wire [18:0]   y_g;
   wire [18:0]   y_b;
   wire [18:0]   u_r;
   wire [18:0]   u_g;
   wire [18:0]   u_b;
   wire [18:0]   v_r;
   wire [18:0]   v_g;
   wire [18:0]   v_b;
   
   wire [18:0]   yuv_y;
   wire [18:0]   yuv_v;
   wire [18:0]   yuv_u;
   
   reg  [18:0]   y_r_reg;
   reg  [18:0]   y_g_reg;
   reg  [18:0]   y_b_reg;
   
   reg  [18:0]   u_r_reg;
   reg  [18:0]   u_g_reg;
   reg  [18:0]   u_b_reg;
   
   reg  [18:0]   v_r_reg;
   reg  [18:0]   v_g_reg;
   reg  [18:0]   v_b_reg;
   
   reg  [18:0]   yuv_y_reg;
   reg  [18:0]   yuv_v_reg;
   reg  [18:0]   yuv_u_reg;

   /*
    * multi for singed number
    */
   function [18:0] mult_tc;
      input [9:0] A;
      input [8:0] B;
      
      reg [9:0]   temp_a;
      reg [8:0]   temp_b;
      reg [17:0]  temp_1;
      reg [17:0]  temp_2;
      reg         sign;
      
      begin
         temp_a = (A[9])? (~A + 1'b1) : A;
         temp_b = (B[8])? (~B + 1'b1) : B;
         
         temp_1 = temp_a * temp_b;
         temp_2 = ~(temp_1 - 1'b1);

         sign = (A[9] ^ B[8]) && (|temp_1);      
         mult_tc = sign ? {1'b1, temp_2} : {1'b0, temp_1}; 
      end
   endfunction // mult_tc

   /*
    * rgb2yuv, 1st stage
    */
   assign y_r = 10'b0100110010 * {1'b0,rgb[7:0]};
   assign y_g = 10'b1001011011 * {1'b0,rgb[15:8]};
   assign y_b = 10'b0001110101 * {1'b0,rgb[23:16]};
   
   assign u_r = mult_tc(10'b1101010011,{1'b0,rgb[7:0]});
   assign u_g = mult_tc(10'b1010101101,{1'b0,rgb[15:8]});
   assign u_b = {2'b0,rgb[23:16],9'b0};
   
   assign v_r = {2'b0,rgb[7:0],9'b0};
   assign v_g = mult_tc(10'b1001010011,{1'b0,rgb[15:8]});
   assign v_b = mult_tc(10'b1110101101,{1'b0,rgb[23:16]});

   always @(posedge HCLK or negedge HRESETn) begin
      if(HRESETn == 1'b0) begin
         y_r_reg <= 19'b0;
         y_g_reg <= 19'b0;
         y_b_reg <= 19'b0;       
         u_r_reg <= 19'b0;
         u_g_reg <= 19'b0;
         u_b_reg <= 19'b0;       
         v_r_reg <= 19'b0;
         v_g_reg <= 19'b0;
         v_b_reg <= 19'b0;
      end
      else if(enable == 1'b1) begin
         y_r_reg <= y_r;
         y_g_reg <= y_g;
         y_b_reg <= y_b;         
         u_r_reg <= u_r;
         u_g_reg <= u_g;
         u_b_reg <= u_b;         
         v_r_reg <= v_r;
         v_g_reg <= v_g;
         v_b_reg <= v_b;
      end
   end
   
   /*
    * reg2yuv, 2nd stage
    */
   assign yuv_y = y_r_reg[18:0] + y_b_reg[18:0] + y_g_reg[18:0];
   assign yuv_u = u_r_reg[18:0] + u_b_reg[18:0] + u_g_reg[18:0];
   assign yuv_v = v_r_reg[18:0] + v_b_reg[18:0] + v_g_reg[18:0];

   always @(posedge HCLK or negedge HRESETn) begin
      if(HRESETn == 1'b0) begin
         yuv_y_reg <= 19'b0;
         yuv_u_reg <= 19'b0;
         yuv_v_reg <= 19'b0;    
      end
      else if(enable == 1'b1) begin
         yuv_y_reg <= yuv_y;
         yuv_u_reg <= yuv_u;
         yuv_v_reg <= yuv_v;     
      end
   end

   /*
    * rgb2yuv, 3rd stage
    */
   assign yuv_out = {yuv_y_reg[17:10], yuv_u_reg[17:10], yuv_v_reg[17:10]};   
   
   always @(posedge HCLK or negedge HRESETn) begin
      if(HRESETn == 1'b0) begin
         yuv <= 24'b0;   
      end
      else if(enable == 1'b1) begin
         yuv <= yuv_out;
      end
   end
   
endmodule
